@@ -7,9 +7,12 @@
 #' @param surv.formula A formula specifying the time-to-event sub-model. Must be usable by 
 #'   \code{\link[survival]{coxph}}.
 #' @param data A \code{data.frame} containing all covariates and responses.
-#' @param family A list of families corresponding in order to \code{long.formula}.
-#' @param post.process Logical, should post processing be done to obtain standard errors and 
-#'   log-likelihood? Defaults to \code{TRUE}.
+#' @param family A list of length \eqn{K} containing strings denoting the exponential families 
+#' for each longitudinal sub-model, corresponding in order to \code{long.formulas}. For choices 
+#' of \code{family}, see \strong{details}.
+#' @param disp.formulas An optional list of length \eqn{K} specifying the dispersion models
+#' wanted for each longitudinal sub-model, corresponding in order to \code{long.formulas}. Defaults
+#' to \code{disp.formulas = NULL}. See \strong{details} for more information.
 #' @param control A list of control values: \describe{
 #' 
 #'   \item{\code{verbose}}{Logical: If \code{TRUE}, at each iteration parameter information will 
@@ -23,25 +26,28 @@
 #'   \strong{details}. Default is \code{tol.den=1e-3}.}
 #'   \item{\code{tol.thr}}{Numeric: Threshold used when \code{conv = 'sas'}, see 
 #'   \strong{details}. Default is \code{tol.thr=1e-1}.}
+#'   \item{\code{maxit}}{Integer: Maximum number of EM iterations to carry out before
+#'   exiting the algorithm. Defaults to \code{maxit=200L}, which is usually sufficient.}
 #'   \item{\code{correlated}}{Logical: Should covariance parameters \strong{between} responses 
 #'   be estimated and used in determination of model convergence? Default is 
 #'   \code{correlated=TRUE}. A choice of \code{correlated=FALSE} is equivalent to imposing the 
 #'   belief that deviations in longitudinal trajectories are not correlated across responses, but
-#'   can \strong{greatly decrease} computation time.}
+#'   can decrease computation time, particularly for large \eqn{K}.}
 #'   \item{\code{gh.nodes}}{Integer: Number of weights and abscissae to use in gauss-hermite 
 #'   quadrature. Defaults to \code{gh.nodes=3}, which is usually sufficient.}
 #'   \item{\code{gh.sigma}}{Numeric: Standard deviation for gauss-hermite approximation of normal
 #'   distribution. Defaults to \code{gh.sigma=1}. This should rarely (if ever) need altering.}
-#'   \item{\code{hessian}}{Character: Determines if the variance-covariance matrix for 
-#'   \eqn{\hat{b}_i}, \eqn{\hat{\Sigma}_i} should be calculated as part of the \code{optim} step
-#'   in minimising the negative log-likelihood, or calculated post-hoc using forward differencing.
-#'   Default is \code{hessian="auto"} for the former, with \code{hessian="manual"} the 
-#'   option for the latter.}
 #'   \item{\code{return.dmats}}{Logical: Should data matrices be returned? Defaults to 
 #'   \code{return.dmats=TRUE}. Note that some S3 methods for \code{\link{joint.object}}s
-#'   greatly benefit from inclusion of these data matrices.}
-#'   \item{\code{center.ph}}{Should the survival covariates be mean-centered? Defaults
+#'   require the returned object to include these data matrices.}
+#'   \item{\code{return.inits}}{Logical: Should a list of inital conditons be returned? 
+#'   Defaults to \code{return.inits=FALSE}.}
+#'   \item{\code{center.ph}}{Logical: Should the survival covariates be mean-centered? Defaults
 #'   to \code{center.ph=TRUE}.}
+#'   \item{\code{post.process}}{Logical: Should model post-processing be carried out (assumes
+#'   that the model has converged). Defaults to \code{post.process = TRUE} which then returns
+#'   posterior modes and their variance for the random effects, as well as approximated standard
+#'   error. This is largely for internal use (i.e. if bootstrapping to obtain SEs instead).}
 #' 
 #' }
 #' 
@@ -79,14 +85,30 @@
 #'     Ismail (2012): \eqn{\varphi>0}: Over-dispersion; \eqn{\varphi<0}: Under-dispersion.
 #'     \eqn{Var[Y]=(1+\varphi)^2\mu}.}
 #'     \item{\code{'Gamma'}}{For continuous data where a Gamma distribution might be sensible.
-#'     The log link is used. A term \eqn{\sigma_k} is be estimated, denoting the shape of the
-#'     distribution.}
+#'     The log link is used. A term \eqn{\sigma_k} is be estimated, denoting the (log) shape of 
+#'     the distribution, which is then reported as \eqn{\varphi_k=\exp\{\sigma_k\}}.}
+#'     \item{\code{"negbin"}}{For count data types where overdispersion is modelled. A log link
+#'     is used. A term \eqn{\sigma_k} is estimated, which is then reported as 
+#'     \eqn{\varphi_k=\exp\{\sigma_k\}} which is the overdispersion. The variance of the response
+#'     is \eqn{Var[Y]=\mu+\mu^2/\varphi}.}
 #'   
 #'   } 
 #'   
-#'   For families where dispersion is estimated, this is \strong{always} specified by an 
-#'   "intercept-only" formula only. This might change in future.
-#'   
+#'  For families \code{"negbin"}, \code{"Gamma"}, \code{"genpois"}, the user can define the 
+#'  dispersion model desired in \code{disp.formulas}. For the \code{"negbin"} and \code{"Gamma"}
+#'  cases, we define \eqn{\varphi_i=\exp\{W_i\sigma_i\}} (i.e. the exponent of the linear 
+#'  predictor of the dispersion model; and for \code{"genpois"} the identity of the linear
+#'  is used.
+#' 
+#' @section Dispersion models:
+#'   The \code{disp.formulas} in the function call allows the user to model the dispersion for
+#'   a given sub-model if wanted. The default value \code{disp.formulas = NULL} simply imposes
+#'   an 'intercept only' model. If the \eqn{k}th item in \code{disp.formulas} corresponds to 
+#'   a longitudinal sub-model with no dispersion term, then it is simply ignored. With this in 
+#'   mind then, if a dispersion model is only required for, say, one sub-model, then the 
+#'   corresponding item in this list of models should be specified as such, with the others set to
+#'   \code{~1}.
+#'
 #' @section Standard error estimation: 
 #'   We follow the approximation of the observed empirical information matrix detailed by 
 #'   Mclachlan and Krishnan (2008), and later used in \code{joineRML} (Hickey et al., 2018).
@@ -134,7 +156,8 @@
 #' 
 #' @seealso \code{\link{summary.joint}}, \code{\link{logLik.joint}}, 
 #' \code{\link{extractAIC.joint}}, \code{\link{fixef.joint}}, \code{\link{ranef.joint}},
-#' \code{\link{vcov.joint}} and \code{\link{joint.object}}.
+#' \code{\link{vcov.joint}}, \code{\link{joint.object}} and \code{\link{xtable.joint}}. For
+#' data simulation see \code{\link{simData}}.
 #' 
 #' @export
 #'
@@ -147,12 +170,12 @@
 #' family <- list('gaussian', 'poisson')
 #' data <- simData(ntms = 10, beta = beta, D = D, n = 100,
 #'                 family = family, zeta = c(0, -0.2),
-#'                 sigma = c(0.16, 0), gamma = gamma)$data
+#'                 sigma = list(0.16, 0), gamma = gamma)$data
 #'
 #' # Specify formulae and target families
 #' long.formulas <- list(
 #'   Y.1 ~ time + cont + bin + (1 + time|id),  # Gaussian
-#'   Y.2 ~ time + cont + bin + (1 + time|id)  # Poisson
+#'   Y.2 ~ time + cont + bin + (1 + time|id)   # Poisson
 #' )
 #' surv.formula <- Surv(survtime, status) ~ bin
 #' 
@@ -161,8 +184,6 @@
 #' \donttest{
 #' # 2) Fit on PBC data -----------------------------------------------------
 #' data(PBC)
-#' PBC$serBilir <- log(PBC$serBilir)
-#'
 #' # Subset data and remove NAs
 #' PBC <- subset(PBC, select = c('id', 'survtime', 'status', 'drug', 'time',
 #'                               'serBilir', 'albumin', 'spiders', 'platelets'))
@@ -170,7 +191,7 @@
 #' 
 #' # Specify GLMM sub-models, including interaction and quadratic time terms
 #' long.formulas <- list(
-#'   serBilir ~ drug * (time + I(time^2)) + (1 + time + I(time^2)|id),
+#'   log(serBilir) ~ drug * (time + I(time^2)) + (1 + time + I(time^2)|id),
 #'   albumin ~ drug * time + (1 + time|id),
 #'   platelets ~ drug * time + (1 + time|id),
 #'   spiders ~ drug * time + (1|id)
@@ -182,104 +203,149 @@
 #'               control = list(verbose = TRUE))
 #' fit
 #' }
-joint <- function(long.formulas, surv.formula, data, family, post.process = TRUE, control = list()){
+#' \donttest{
+#' # 3) Fit with dispersion models ----------------------------------------
+#' beta <- do.call(rbind, replicate(2, c(2, -0.1, 0.1, -0.2), simplify = FALSE))
+#' gamma <- c(0.3, -0.3)
+#' D <- diag(c(0.25, 0.09, 0.25, 0.05))
+#' family <- list('negbin', 'poisson')   # As an example; only requires one dispersion model.
+#' sigma <- list(c(1, 0.2), 0)           # Need to specify the model in simData call too.
+#' disp.formulas = list(~time, ~1)       # Even though poisson doesn't model dispersion, need to
+#'                                       # populate this element in disp.formulas!
+#' # Simulate some data
+#' data <- simData(ntms = 10, beta = beta, D = D, n = 500,
+#'                 family = family, zeta = c(0, -0.2), sigma = sigma,
+#'                 disp.formulas = disp.formulas, gamma = gamma)$data
+#'
+#' # Now fit using joint
+#' long.formulas <- list(
+#'   Y.1 ~ time + cont + bin + (1+time|id),
+#'   Y.2 ~ time + cont + bin + (1+time|id)
+#' )
+#' fit <- joint(
+#'   long.formulas, Surv(survtime, status) ~ bin,
+#'   data, family, disp.formulas = disp.formulas
+#' )
+#' fit
+#' summary(fit)
+#' }
+joint <- function(long.formulas, surv.formula, 
+                  data, family,
+                  disp.formulas = NULL, 
+                  control = list()){
+  
+  con <- list(correlated = T, gh.nodes = 3, gh.sigma = 1, center.ph = T,
+              tol.abs = 1e-3, tol.rel = 1e-2, tol.thr = 1e-1, tol.den = 1e-3,
+              maxit = 200, conv = 'sas', verbose = F, return.inits = F, return.dmats = T, post.process = T)
+  conname <- names(con)
+  if(any(!names(control)%in%conname)){
+    warning("Supplied control arguments do not match with possible names:\n", 
+            paste0(sapply(conname, sQuote), collapse=', '), '.')
+  }
+  con[(conname <- names(control))] <- control
+  
+  # Ensure supplied families are character, not functions
+  if(!is.list(family) | !all(sapply(family, function(x) is.character(x) & !is.function(x))))
+    stop(sQuote("family"), " must be supplied as a list of character strings")
   
   start.time <- proc.time()[3]
   
   # Initial parsing ----
+  pf <- parent.frame()
   if("factor"%in%class(data$id)) data$id <- as.numeric(as.character(data$id))
   formulas <- lapply(long.formulas, parseFormula)
-  center.ph <- if(!is.null(control$center.ph)) control$center.ph else TRUE
-  surv <- parseCoxph(surv.formula, data, center.ph)
+  surv <- parseCoxph(surv.formula, data, con$center.ph)
   n <- surv$n; K <- length(family)
-  if(K!=length(long.formulas)) stop('Mismatched lengths of "family" and "long.formulas".')
+  if(K!=length(long.formulas)) stop('Mismatched lengths of ', sQuote("family"), " and ", sQuote("long.formulas"),'.')
+  # Parse dispersion formulas
+  if(is.null(disp.formulas)){
+    disp.formulas <- replicate(K, ~1, simplify = FALSE)
+  }else{
+    if(sum(!sapply(disp.formulas, is.null)) != K)
+      stop("Need to supply dispersion formulas for all responses even if not required for all K responses\n",
+           "(You can just fill-in ", sQuote("~1"), " for those with no wanted dispersion model.)")
+  }
+  # Reduce memory overheads, particularly for returned object(s) ?
+  disp.formulas <- lapply(disp.formulas, function(x){
+    environment(x) <- pf
+    x
+  })
   
   # Initial conditons ----
-  inits.long <- Longit.inits(long.formulas, data, family)
+  inits.long <- Longit.inits(long.formulas, disp.formulas, data, family)
+  id.assign <- AssignIds(data)
+  dmats <- getdmats(inits.long$fits, id.assign)
   # Suss out indices of b_k and beta_k.
-  b.inds <- lapply(1:K, function(k){
-    nm <- inits.long$responses[k]
-    which(grepl(nm, colnames(inits.long$b)))
+  b.inds <- lapply(seq_along(dmats$q), function(x){
+    xx <- seq(dmats$q[x])
+    if(x > 1) 
+      return(xx + sum(dmats$q[1:(x-1)]))
+    else
+      return(xx)
   })
-  beta.inds <- lapply(1:K, function(k){
-    nm <- inits.long$responses[k]
-    which(grepl(nm, names(inits.long$beta.init)))
+  beta.inds <- lapply(seq_along(dmats$P), function(x){
+    xx <- seq(dmats$P[x])
+    if(x > 1) 
+      return(xx + sum(dmats$P[1:(x-1)]))
+    else
+      return(xx)
   })
-  q <- length(do.call(c, b.inds))
+  inds <- list(
+    R = list(b = b.inds, beta = beta.inds),
+    Cpp = list(b = lapply(b.inds, function(x) x - 1), beta = lapply(beta.inds, function(x) x - 1))
+  )
   
-  inits.surv <- TimeVarCox(data, inits.long$b, surv, formulas, b.inds)
+  inits.surv <- TimeVarCox(data, inits.long$b, surv, formulas, b.inds, inits.long)
   
   # Longitudinal parameters
   beta <- inits.long$beta.init
+  names.beta <- names(beta)
   D <- inits.long$D.init
-  sigma <- inits.long$sigma.init # dispersion / resid. variance / 0 otherwise.
+  sigma <- inits.long$sigma.init # dispersions
   b <- lapply(1:n, function(i) inits.long$b[i, ])
   # Survival parameters
   zeta <- inits.surv$inits[match(names(surv$ph$assign), names(inits.surv$inits))]
   names(zeta) <- paste0('zeta_', names(zeta))
   gamma <- inits.surv$inits[grepl('gamma\\_', names(inits.surv$inits))]
   
-  # Longitudinal and survival data objects ----
-  dmats <- createDataMatrices(data, formulas)
-  sv <- surv.mod(surv, formulas, inits.surv$l0.init)
-  
-  X <- dmats$X; Y <- dmats$Y; Z <- dmats$Z # Longitudinal data matrices
-  m <- lapply(Y, function(y) sapply(y, length))
-  # survival
-  Fi <- sv$Fi; Fu <- sv$Fu; l0i <- sv$l0i; l0u <- sv$l0u; Delta <- surv$Delta 
-  l0 <- sv$l0
-  S <- sv$S; SS <- sv$SS
-  
-  # Assign family to joint density and parameter updates ----
-  # Ensure family is a string, not a function i.e. if user supplies `Gamma` instead of `"Gamma"`.
-  family <- lapply(family, function(family) if("function"%in%class(family)) family()$family else family)
-  
-  # Do we want to add-in correlated random effects between responses? For large K this greatly increases
-  #   computation time and model instability.
-  if(!is.null(control$correlated)) correlated <- control$correlated else correlated <- T
-  if(!correlated) D[inits.long$off.inds] <- 0
+  # Survival data objects 
+  sv <- surv.mod(surv, formulas, inits.surv$l0.init, inits.long)
   
   # Parameter vector and list ----
   Omega <- list(D=D, beta = beta, sigma = sigma, gamma = gamma, zeta = zeta)
   params <- c(setNames(vech(D), paste0('D[', apply(which(lower.tri(D, T), arr.ind = T), 1, paste, collapse = ','), ']')),
               beta, unlist(sigma)[inits.long$sigma.include], gamma, zeta)
-
+  sigma.include <- inits.long$sigma.include
+  if(!con$return.inits) rm(inits.surv)
+  
   # Gauss-Hermite Quadrature ----
-  if(!is.null(control$gh.nodes)) gh <- control$gh.nodes else gh <- 3
-  if(!is.null(control$gh.sigma)) .sigma <- control$gh.sigma else .sigma <- 1
-  
-  GH <- statmod::gauss.quad.prob(gh, 'normal', sigma = .sigma)
+  GH <- statmod::gauss.quad.prob(con$gh.nodes, 'normal', sigma = con$gh.sigma)
   w <- GH$w; v <- GH$n
-  
+
   # Begin EM ----
   diff <- 100; iter <- 0;
   # Convergence criteria setup
-  if(!is.null(control$tol.abs)) tol.abs <- control$tol.abs else tol.abs <- 1e-3
-  if(!is.null(control$tol.rel)) tol.rel <- control$tol.rel else tol.rel <- 1e-2
-  if(!is.null(control$tol.den)) tol.den <- control$tol.den else tol.den <- 1e-3
-  if(!is.null(control$tol.thr)) tol.thr <- control$tol.thr else tol.thr <- 1e-1
-  if(!is.null(control$maxit)) maxit <- control$maxit else maxit <- 200
-  if(!is.null(control$conv)) conv <- control$conv else conv <- "sas"
-  if(!conv%in%c('absolute', 'relative', 'either', 'sas'))
-    stop('Convergence criteria must be one of "absolute", "relative", "either" or "sas".')
-  convergence.criteria <- list(type = conv, tol.abs = tol.abs, tol.rel = tol.rel, tol.den = tol.den,
-                               threshold = tol.thr)
+  if(!con$conv%in%c('absolute', 'relative', 'either', 'sas')){
+    warning('Convergence criteria must be one of "absolute", "relative", "either" or "sas". Using "sas"')
+    con$conv <- "sas"
+  }
+    
+  convergence.criteria <- list(type = con$conv, tol.abs = con$tol.abs, tol.rel = con$tol.rel, tol.den = con$tol.den,
+                               threshold = con$tol.thr)
   
-  if(!is.null(control$verbose)) verbose <- control$verbose else verbose <- F
-  if(!is.null(control$hessian)) hessian <- control$hessian else hessian <- 'auto'
-  if(!hessian %in% c('auto', 'manual')) stop("Argument 'hessian' needs to be either 'auto' (i.e. from optim) or 'manual' (i.e. from _sdb, the default).")
-  if(!is.null(control$return.inits)) return.inits <- control$return.inits else return.inits <- F
-  if(!is.null(control$return.dmats)) return.dmats <- control$return.dmats else return.dmats <- T
-  
-  if(verbose) cat("Starting EM algorithm...\n")
+  if(con$verbose){
+    cat("Initial conditions: \n")
+    converge.check(0, 0, convergence.criteria, 0, Omega, TRUE)
+    cat("\nStarting EM algorithm...\n")
+  }
   converged <- FALSE
   EMstart <- proc.time()[3]
-  while((!converged) && (iter < maxit)){
-    update <- EMupdate(Omega, family, X, Y, Z, b, 
-                       S, SS, Fi, Fu, l0i, l0u, Delta, l0, sv, 
-                       w, v, n, m, hessian, beta.inds, b.inds, K, q)
-    if(!correlated) update$D[inits.long$off.inds] <- 0
-    params.new <- c(vech(update$D), update$beta, unlist(update$sigma)[inits.long$sigma.include], 
+  while((!converged) && (iter < con$maxit)){
+    update <- EMupdate(Omega, family, dmats, b, sv, 
+                       surv, w, v, con, inds)
+    
+    if(!con$correlated) update$D[inits.long$off.inds] <- 0
+    params.new <- c(vech(update$D), update$beta, unlist(update$sigma)[sigma.include], 
                     update$gamma, update$zeta)
     names(params.new) <- names(params)
     
@@ -288,17 +354,21 @@ joint <- function(long.formulas, surv.formula, data, family, post.process = TRUE
     D <- update$D; beta <- update$beta; sigma <- update$sigma
     gamma <- update$gamma; zeta <- update$zeta
     l0 <- update$l0; l0u <- update$l0u; l0i <- update$l0i
+    sv$l0 <- l0; sv$l0i <- l0i; sv$l0u <- l0u
     iter <- iter + 1
     Omega <- list(D = D, beta = beta, sigma = sigma, gamma = gamma, zeta = zeta)
     
-    convcheck <- converge.check(params, params.new, convergence.criteria, iter, Omega, verbose)
+    convcheck <- converge.check(params, params.new, convergence.criteria, iter, Omega, con$verbose)
     if(iter >= 4) converged <- convcheck$converged # Allow to converge after 3 iterations
     params <- params.new
   }
   
   EMend <- proc.time()[3]
   coeffs <- Omega
-  coeffs$beta <- setNames(c(Omega$beta), names(inits.long$beta.init))
+  coeffs$beta <- setNames(c(Omega$beta), names.beta)
+  if(is.not.SPD(coeffs$D)) warning("Covariance matrix D is not positive semi-definite at convergence,",
+                                   " potential model misspecification or lower tolerance options required.")
+  
   out <- list(coeffs = coeffs,
               hazard = cbind(ft = sv$ft, haz = l0, nev = sv$nev))
   
@@ -306,48 +376,35 @@ joint <- function(long.formulas, surv.formula, data, family, post.process = TRUE
   ModelInfo$ResponseInfo <- sapply(1:K, function(k){
     paste0(inits.long$responses[k], ' (', family[k], ')')
   })
+  ModelInfo$Resps <- inits.long$responses
   ModelInfo$family <- family
+  ModelInfo$K <- dmats$K
+  ModelInfo$Pcounts <- list(P = dmats$P, Pd = dmats$Pd, 
+                            q = sv$q, vD = length(vech(D)))
   ModelInfo$long.formulas <- long.formulas
-  ModelInfo$surv.formulas <- surv.formula
+  ModelInfo$disp.formulas <- disp.formulas
+  ModelInfo$surv.formula <- surv.formula
   ModelInfo$survtime <- surv$survtime
   ModelInfo$status <- surv$status
-  ModelInfo$control <- if(!is.null(control)) control else NULL
+  ModelInfo$control <- con
   ModelInfo$convergence.criteria <- convergence.criteria
-  ModelInfo$inds <- list(beta = beta.inds, b = b.inds)
-  ModelInfo$nobs <- colSums(do.call(rbind, m))
-  ModelInfo$n <- n
+  ModelInfo$inds <- inds
+  ModelInfo$n <- dmats$n
+  ModelInfo$nobs <- setNames(dmats$m, inits.long$responses)
+  ModelInfo$mi <- sapply(dmats$mi, unlist)
   ModelInfo$nev <- sum(sv$nev)
+  ModelInfo$id.assign <- list(original.id = id.assign$id,
+                              assigned.id = id.assign$assign)
   out$ModelInfo <- ModelInfo
   
   # Post processing ----
-  if(post.process){
-    if(verbose) cat('Calculating SEs\n')
-    beta.inds2 <- lapply(beta.inds, function(x) x - 1); b.inds2 <- lapply(b.inds, function(x) x - 1) 
+  if(con$post.process){
+    if(con$verbose) cat('Post-processing...\n')
     gamma.rep <- rep(gamma, sapply(b.inds, length))
     pp.start.time <- proc.time()[3]
     
-    # Evaluate l0 at final parameter estimates and obtain l0u from surv.mod.
-    l0 <- sv$nev/rowSums(lambdaUpdate(sv$surv.times2, sv$ft.mat, gamma, zeta, sv$S, update$Sigma, b, w, v, b.inds2))
-    sv.new <- surv.mod(surv, formulas, l0)
-    
-    # b and Sigmai at MLEs
-    b.update <- mapply(function(b, Y, X, Z, Delta, S, Fi, l0i, SS, Fu, l0u){
-      optim(b, joint_density, joint_density_ddb,
-            Y = Y, X = X, Z = Z, beta = beta, D = D, sigma = sigma, family = family, 
-            Delta = Delta, S = S, Fi = Fi, l0i = l0i, SS = SS, Fu = Fu, haz = l0u, gamma_rep = gamma.rep, zeta = zeta,
-            beta_inds = beta.inds2, b_inds = b.inds2, K = K,
-            method = 'BFGS', hessian = T)
-    }, b = b, Y = Y, X = X, Z = Z, Delta = Delta, S = S, Fi = Fi, l0i = sv.new$l0i, SS = SS,
-    Fu = Fu, l0u = sv.new$l0u, SIMPLIFY = F)
-    Sigma <- lapply(b.update, function(x) solve(x$hessian))
-    b <- lapply(b.update, function(x) x$par)
-    SigmaSplit <- lapply(Sigma, function(x) lapply(b.inds, function(y) as.matrix(x[y,y])))
-    bsplit <- lapply(b, function(x) lapply(b.inds, function(y) x[y]))
-
-    # The Information matrix
-    II <- obs.emp.I(coeffs, dmats, surv, sv, 
-                    Sigma, SigmaSplit, b, bsplit, 
-                    sv.new$l0u, w, v, n, family, K, q, beta.inds, b.inds)
+    II <- obs.emp.I(coeffs, dmats, surv, sv, family, b, 
+                    l0i, l0u, w, v, inds)
     H <- structure(II$Hessian,
                    dimnames = list(names(params), names(params)))
     
@@ -357,26 +414,32 @@ joint <- function(long.formulas, surv.formula, data, family, post.process = TRUE
     out$vcov <- I
     out$SE <- sqrt(diag(I))
    
-    postprocess.time <- round(proc.time()[3]-pp.start.time, 2)
+    postprocess.time <- round(proc.time()[3] - pp.start.time, 2)
+    
     # Calculate log-likelihood. Done separately as EMtime + postprocess.time is for EM + SEs.
-    out$logLik <- joint.log.lik(coeffs, dmats, b, surv, sv, sv.new$l0u, sv.new$l0i, gamma.rep, beta.inds, b.inds, 
-                                K, q, family, Sigma)
+    out$logLik <- joint.log.lik(coeffs, dmats, surv, sv, family, II$b.hat, l0i, l0u, inds, II$Sigma)
     # Collate RE and their variance
+    REs <- do.call(rbind, II$b.hat)
+    attr(REs, 'Var') <- do.call(rbind, lapply(II$Sigma, diag))
+    attr(REs, 'vcov') <- do.call(rbind, lapply(II$Sigma, vech))
+    out$REs <- REs
+  }else{
     REs <- do.call(rbind, b)
-    attr(REs, 'Var') <- do.call(rbind, lapply(Sigma, diag))
-    attr(REs, 'vcov') <- do.call(rbind, lapply(Sigma, vech))
+    attr(REs, 'Var') <- do.call(rbind, lapply(update$Sigma, diag))
+    attr(REs, 'vcov') <- do.call(rbind, lapply(update$Sigma, vech))
     out$REs <- REs
   }
   comp.time <- round(proc.time()[3] - start.time, 3)
   out$elapsed.time <- c(`EM time` = unname(round(EMend - EMstart, 3)),
-                        `Post processing` = if(post.process) unname(postprocess.time) else NULL,
+                        `Post processing` = if(con$post.process) unname(postprocess.time) else NULL,
                         `Total Computation time` = unname(comp.time),
                         `iterations` = iter)
+  if(con$post.process) sv <- surv.mod(surv, formulas, l0, inits.long)
+  dmats <- list(long = dmats, surv = sv, ph = surv)
+  if(con$return.dmats) out$dmats <- dmats
   
-  dmats <- list(long = dmats, surv = if(post.process) sv.new else sv, ph = surv)
-  if(return.dmats) out$dmats <- dmats
-  
-  if(return.inits) out$inits = inits.long
+  if(con$return.inits) out$inits = list(inits.long = inits.long,
+                                        inits.surv = inits.surv)
   class(out) <- 'joint'
   return(out)
 }
@@ -388,7 +451,11 @@ print.joint <- function(x, ...){
   if(!inherits(x, 'joint')) stop('x must be a "joint" object!')
   
   M <- x$ModelInfo
-  K <- length(M$ResponseInfo) # Number of responses
+  K <- M$K # Number of responses
+  fams <- M$family
+  dpsL <- sapply(M$long.formulas, long.formula.to.print, 1)
+  dpsD <- sapply(M$disp.formulas, deparse)
+  dpsS <- deparse(M$surv.formula)
   
   # Data information
   cat(sprintf("Number of subjects: %d\n", M$n))
@@ -396,22 +463,25 @@ print.joint <- function(x, ...){
   
   # Longitudinal information: 
   cat("\n===================\nModel specification\n===================\n")
-  if(K == 1){
+  if(K == 1)
     cat("Univariate longitudinal process specification:\n")
-    cat(sprintf("%s: %s\n", M$ResponseInfo[1], deparse(M$long.formulas[[1]])))
-  }else{
+  else
     cat("Multivariate longitudinal process specifications: \n")
-    for(k in 1:K){
-      cat(sprintf("%s: %s\n", M$ResponseInfo[k], deparse(M$long.formulas[[k]])))
-    }
+  
+  for(k in 1:K){
+    cat(sprintf("%s: %s\n", M$ResponseInfo[k], dpsL[k]))
+    if(fams[[k]] %in% c("Gamma", "negbin", "genpois"))
+      cat(sprintf("Dispersion model: %s\n",
+                  if(dpsD[k]=="~1") "(Intercept) model" else dpsD[k]
+      ))
   }
   
   cat("\nSurvival sub-model specification: \n")
-  cat(deparse(M$surv.formulas))
+  cat(deparse(M$surv.formula))
   
   cat("\n\nAssociation parameter estimates: \n")
   print(setNames(x$coeffs$gamma,
-        M$ResponseInfo))
+        M$Resps))
   cat("\n")
-  invisible(1+1)
+  invisible(x)
 }
