@@ -1,10 +1,24 @@
+#' @keywords internal
+dgenpois <- function(x, mu, phi, log = FALSE){
+  lx <- length(x); lm <- length(mu); lp <- length(phi)
+  ch <- lx == lm & lx == lp & lm == lp
+  if(!ch) stop("Not all supplied inputs are of same length, check.\n")
+  
+  out <- sapply(1:lx, function(i) GP1_pmf_scalar(mu[i], phi[i], x[i]))
+  if(log)
+    return(log(out))
+  else
+    return(out)
+}
+
 #' Simulate realisations from a generalised poisson distribution
 #' @param mu A numeric vector of rates \eqn{\exp{\eta}}, with \eqn{\eta} the linear predictor.
 #' @param phi A numeric specifying the dispersion \eqn{\varphi}. If \eqn{\varphi<0} the response 
 #' will be under-dispersed and overdispersed if \eqn{\varphi>0}.
 #' 
 #' @details Follows the "GP-1" implementation of the generalised Poisson distribution outlined 
-#' in Zamani & Ismail (2012). The variance of produced \eqn{Y} is \eqn{(1+\varphi)^2\mu}.
+#' in Zamani & Ismail (2012). The variance of produced \eqn{Y} is \eqn{(1+\varphi)^2\mu}. As such
+#' the dispersion parameter is bounded (i.e. not in positive reals as with CMP distribution).
 #' 
 #' @return An appropriately-dimensioned vector of count data.
 #' 
@@ -20,11 +34,12 @@ rgenpois <- function(mu, phi){
   for(j in 1:n){
     ans <- 0;
     rand <- runif(1)
-    kum <- GP1_pmf_scalar(mu[j], phi, 0)
+    kum <- dgenpois(0, mu[j], phi[j])
     while(rand > kum){
       ans <- ans + 1
-      # message(ans)
-      kum <- kum + GP1_pmf_scalar(mu[j], phi, ans)
+      kum <- kum + dgenpois(ans, mu[j], phi[j])
+      # For debugging
+      if(is.infinite(kum) || is.nan(kum)) cat(sprintf("NaN/Infinite kum: mu = %.2f, phi = %.2f", mu[j], phi[j]))
     }
     out[j] <- ans
   }
@@ -42,9 +57,9 @@ rgenpois <- function(mu, phi){
 #' @param fup the maximum follow-up time, such that t = [0, ..., fup] with length \code{ntms}. 
 #' In instances where subject \eqn{i} \emph{doesn't} fail before \code{fup}, their censoring
 #' time is set as \code{fup + 0.1}.
-#' @param family a \eqn{K}-list of families, see \strong{details}.
-#' @param sigma a \eqn{K}-vector of dispersion parameters corresponding to the order of 
-#' \code{family}; see \strong{details}.
+#' @param family a \eqn{K}-\code{list} of families, see \strong{details}.
+#' @param sigma a \eqn{K}-\code{list} of dispersion parameters corresponding to the order of 
+#' \code{family}, and matching \code{disp.formulas} specification; see \strong{details}.
 #' @param beta a \eqn{K \times 4} matrix specifying fixed effects for each \eqn{K} parameter, 
 #' in the order (Intercept), time, continuous, binary.
 #' @param D a positive-definite matrix specifying the variance-covariance matrix for the random
@@ -55,23 +70,34 @@ rgenpois <- function(mu, phi){
 #' the survival sub-model, in the order of continuous and binary.
 #' @param theta parameters to control the failure rate, see \strong{baseline hazard}.
 #' @param cens.rate parameter for \code{rexp} to generate censoring times for each subject.
-#' @param unif.times logical, if \code{unif.times = TRUE} (the default), then \emph{every} subject
-#' will have the same follow-up times defined by \code{seq(0, fup, length.out = ntms)}. If 
-#' \code{unif.times = FALSE} then follow-up times are set as random draws from a uniform 
-#' distribution with maximum \code{fup}. 
-#' @param random.formula allows user to specify if an intercept-and-slope (\code{~ time}) or 
-#' intercept-only (\code{~1}) random effects structure should be used. defaults to the former.
+#' @param regular.times logical, if \code{regular.times = TRUE} (the default), then 
+#' \emph{every} subject will have the same follow-up times defined by 
+#' \code{seq(0, fup, length.out = ntms)}. If \code{regular.times = FALSE} then follow-up times are
+#' set as random draws from a uniform distribution with maximum \code{fup}. 
+#' @param dof integer, specifies the degrees of freedom of the multivariate t-distribution
+#' used to generate the random effects. If specified, this t-distribution is used. If left
+#' at the default \code{dof=Inf} then the random effects are drawn from a multivariate normal
+#' distribution.
+#' @param random.formulas allows user to specify if an intercept-and-slope (\code{~ time}) or 
+#' intercept-only (\code{~1}) random effects structure should be used on a response-by-response
+#' basis. Defaults to an intercept-and-slope for all responses.
+#' @param disp.formulas allows user to specify the dispersion model simulated. Intended use is
+#' to allow swapping between intercept only (the default) and a time-varying one (\code{~ time}).
+#' Note that this should be a \eqn{K}-\code{list} of formula objects, so if only one dispersion 
+#' model is wanted, then an intercept-only should be specified for remaining sub-models. The
+#' corresponding item in list of \code{sigma} parameters should be of appropriate size. Defaults
+#' to an intercept-only model.
 #' @param return.ranefs a logical determining whether the \emph{true} random effects should be 
 #' returned. This is largely for internal/simulation use. Default \code{return.ranefs = FALSE}.
-#'   
+#'
 #' @returns A list of two \code{data.frame}s: One with the full longitudinal data, and another 
 #' with only survival data. If \code{return.ranefs=TRUE}, a matrix of the true \eqn{b} values is
-#' also returned.
+#' also returned. By default (i.e. no arguments provided), a bivariate Gaussian set of joint
+#' data is returned.
 #'
 #' @details \code{simData} simulates data from a multivariate joint model with a mixture of 
-#' families for each \eqn{K=1,\dots,3} response. Currently, the argument \code{random.formula}
-#' specifies the association structure for \strong{all} responses. The specification of 
-#' \code{family} changes requisite dispersion parameter, if applicable. The \code{family} list can
+#' families for each \eqn{k=1,\dots,K} response. The specification of \code{family} changes
+#' requisite dispersion parameter \code{sigma}, if applicable. The \code{family} list can
 #' (currently) contain: 
 #'   
 #'   \describe{
@@ -82,15 +108,24 @@ rgenpois <- function(mu, phi){
 #'   can be anything, as it doesn't impact simulation.}
 #'   \item{\code{"binomial"}}{Simulated with logit link, corresponding dispersion in \code{sigma} 
 #'   can be anything, as it doesn't impact simulation.}
+#'   \item{\code{"negbin"}}{Simulated with a log link, corresponding item in \code{sigma} will be
+#'   the \strong{overdispersion} defined on the log scale. Simulated variance is 
+#'   \eqn{\mu+\mu^2/\varphi}.}
 #'   \item{\code{"genpois"}}{Simulated with a log link, corresponding item in \code{sigma} will be
 #'   the \strong{dispersion}. Values < 0 correspond to under-dispersion, and values > 0 over-
 #'   dispersion. See \code{\link{rgenpois}} for more information. Simulated variance is 
 #'   \eqn{(1+\varphi)^2\mu}.}
 #'   \item{\code{"Gamma"}}{Simulated with a log link, corresponding item in \code{sigma} will be
-#'   the \strong{shape}.}
+#'   the \strong{shape} parameter, defined on the log-scale.}
 #'   
 #'   }
 #'   
+#'  Therefore, for families \code{"negbin"}, \code{"Gamma"}, \code{"genpois"}, the user can
+#'  define the dispersion model desired in \code{disp.formulas}, which creates a data matrix 
+#'  \eqn{W}. For the \code{"negbin"} and \code{"Gamma"} cases, we define 
+#'  \eqn{\varphi_i=\exp\{W_i\sigma_i\}} (i.e. the exponent of the linear predictor of the 
+#'  dispersion model); and for \code{"genpois"} the identity of the linear is used.
+#'   
 #' @section Baseline hazard: 
 #'  
 #'   When simulating the survival time, the baseline hazard is a Gompertz distribution controlled 
@@ -101,9 +136,11 @@ rgenpois <- function(mu, phi){
 #'   where \eqn{y} is the shape parameter, and the scale parameter is \eqn{\exp{x}}.
 #'
 #' @importFrom MASS mvrnorm
+#' @importFrom mvtnorm rmvt
 #' 
 #' @author James Murray (\email{j.murray7@@ncl.ac.uk}).
 #' @keywords simulation
+#' @seealso \code{\link{joint}}
 #' 
 #' @references 
 #' 
@@ -114,52 +151,86 @@ rgenpois <- function(mu, phi){
 #' @export 
 #'
 #' @examples
+#' # 1) A set of univariate data ------------------------------------------
+#' beta <- c(2.0, 0.33, -0.25, 0.15)
+#' # Note that by default arguments are bivariate, so need to specify the univariate case
+#' univ.data <- simData(beta = beta,    
+#'                      gamma = 0.15, sigma = list(0.2), family = list("gaussian"), 
+#'                      D = diag(c(0.25, 0.05)))
+#'                      
+#' # 2) Univariate data, with failure rate controlled ---------------------
+#' # In reality, many facets contribute to the simulated failure rate, in 
+#' # this example, we'll just atler the baseline hazard via 'theta'.
+#' univ.data.highfail <- simData(beta = beta,
+#'                               gamma = 0.15, sigma = list(0.0), family = list("poisson"),
+#'                               D = diag(c(0.40, 0.08)), theta = c(-2, 0.1))
 #' 
-#' # K = 3 mixture of families with dispersion parameters
+#' # 3) Trivariate (K = 3) mixture of families with dispersion parameters -
 #' beta <- do.call(rbind, replicate(3, c(2, -0.1, 0.1, -0.2), simplify = FALSE))
 #' gamma <- c(0.3, -0.3, 0.3)
 #' D <- diag(c(0.25, 0.09, 0.25, 0.05, 0.25, 0.09))
-#' family <- list('gaussian', 'genpois', 'Gamma')
-#' sigma <- c(.16, 1.5, 1.5)
-#' sim.data <- simData(ntms=15, family = family, sigma = sigma, beta = beta, D = D, gamma = gamma,
-#'                     theta = c(-3, 0.2), zeta = c(0,-.2))
+#' family <- list('gaussian', 'genpois', 'negbin')
+#' sigma <- list(.16, 1.5, log(1.5))
+#' triv.data <- simData(ntms=15, family = family, sigma = sigma, beta = beta, D = D, 
+#'                      gamma = gamma, theta = c(-3, 0.2), zeta = c(0,-.2))
 #' 
-#' # K = 4 mixture of families with/out dispersion
+#' # 4) K = 4 mixture of families with/out dispersion ---------------------
 #' beta <- do.call(rbind, replicate(4, c(2, -0.1, 0.1, -0.2), simplify = FALSE))
 #' gamma <- c(-0.75, 0.3, -0.6, 0.5)
 #' D <- diag(c(0.25, 0.09, 0.25, 0.05, 0.25, 0.09, 0.16, 0.02))
 #' family <- list('gaussian', 'poisson', 'binomial', 'gaussian')
-#' sigma <- c(.16, 0, 0, .05)
-#' sim.data <- simData(ntms=15, family = family, sigma = sigma, beta = beta, D = D, gamma = gamma,
+#' sigma <- list(.16, 0, 0, .05) # 0 can be anything here, as it is ignored internally.
+#' mix.data <- simData(ntms=15, family = family, sigma = sigma, beta = beta, D = D, gamma = gamma,
 #'                     theta = c(-3, 0.2), zeta = c(0,-.2))
+#'                     
+#' # 5) Bivariate joint model with two dispersion models. -----------------
+#' disp.formulas <- list(~time, ~time)          # Two time-varying dispersion models
+#' sigma <- list(c(0.00, -0.10), c(0.10, 0.15)) # specified in form of intercept, slope
+#' D <- diag(c(.25, 0.04, 0.50, 0.10))
+#' disp.data <- simData(family = list("genpois", "negbin"), sigma = sigma, D = D,
+#'                      beta = rbind(c(0, 0.05, -0.15, 0.00), 1 + c(0, 0.25, 0.15, -0.20)),
+#'                      gamma = c(1.5, 1.5),
+#'                      disp.formulas = disp.formulas, fup = 5)            
+#'
+#' # 6) Trivariate joint model with mixture of random effects models ------
+#' # It can be hard to e.g. fit a binomial model on an intercept and slope, since e.g.
+#' # glmmTMB might struggle to accurately fit it (singular fits, etc.). To that end, could
+#' # swap the corresponding random effects specification to be an intercept-only.
+#' family <- list("gaussian", "binomial", "gaussian")
+#' # A list of formulae, even though we want to change the second sub-model's specification
+#' # we need to specify the rest of the items, too (same as disp.formulas, sigma).
+#' random.formulas <- list(~time, ~1, ~time)
+#' beta <- rbind(c(2, -0.2, 0.5, -0.25), c(0, 0.5, 1, -1), c(-2, 0.2, -0.5, 0.25))
+#' # NOTE that the specification of RE matrix will need to match.
+#' D <- diag(c(0.25, 0.09, 1, 0.33, 0.05))
+#' # Simulate data, and return REs as a sanity check...
+#' mix.REspec.data <- simData(beta = beta, D = D, family = family,
+#'                            gamma = c(-0.5, 1, 0.5), sigma = list(0.15, 0, 0.15),
+#'                            random.formulas = random.formulas, return.ranefs = TRUE)
 simData <- function(n = 250, ntms = 10, fup = 5, 
                     family = list('gaussian', 'gaussian'), 
-                    sigma = c(0.16, 0.16),
+                    sigma = list(0.16, 0.16),
                     beta = rbind(c(1, 0.10, 0.33, -0.50), c(1, 0.10, 0.33, -0.50)), D = NULL,
                     gamma = c(0.5, -0.5), zeta = c(0.05, -0.30),
                     theta = c(-4, 0.2), cens.rate = exp(-3.5),
-                    unif.times = TRUE,
-                    random.formula = NULL,
+                    regular.times = TRUE,
+                    dof = Inf,
+                    random.formulas = NULL, disp.formulas = NULL,
                     return.ranefs = FALSE){
-  
+  random.formula <- random.formulas
   # Checks --------------
   # Check family is valid option
   family <- lapply(family, function(f){
     if("function"%in%class(f)) f <- f()$family
-    if(!f%in%c("gaussian", "binomial", "poisson", "genpois", "Gamma")) 
-      stop('Family must be one of "gaussian", "binomial", "poisson", "genpois" or "Gamma".')
+    if(!f%in%c('gaussian', 'binomial', 'poisson', 'genpois', 'Gamma', 'genpois', 'negbin')) 
+      stop('Family must be one of "gaussian", "binomial", "poisson", "genpois", "negbin" or "Gamma".')
     f
   })
   # Check dispersion supplied if neg. binom.
-  family <- lapply(family, function(f) family <- match.arg(f, c("gaussian", "binomial", "poisson", "genpois", "Gamma"), several.ok = F))
-  
-  # Checks wrt the fixed effects and dispersion parameters
+  family <- lapply(family, function(f) match.arg(f, c('gaussian', 'binomial', 'poisson', 'genpois', 'Gamma', 'genpois', 'negbin'), several.ok = F))
   funlist <- unlist(family)
-  num.gp <- length(which(funlist == 'genpois'))
-  num.ga <- length(which(funlist == 'gaussian'))
-  num.Ga <- length(which(funlist == 'Gamma'))
-  
   # Fixed effects
+  if(!"matrix"%in%class(beta)) beta <- t(beta) # Assume if a vector is passed as beta, then a univariate set of data is being attempted.
   K <- nrow(beta)
   if(K != length(funlist)) stop('Incorrect number of families provided and/or beta terms.')
   if(K != length(gamma)) stop('Incorrect number of association parameters provided wrt beta terms.')
@@ -171,10 +242,21 @@ simData <- function(n = 250, ntms = 10, fup = 5,
   }
   
   # Check covariance matrix D is positive semi-definite.
-  if(any(eigen(D)$value < 0) || det(D) <= 0) stop('D must be positive semi-definite')
+  if(is.not.SPD(D)) stop('D must be positive semi-definite')
+  
+  # Parse dispersion formulas
+  pf <- parent.frame()
+  if(is.null(disp.formulas)){
+    disp.formulas <- replicate(K, ~1, simplify = FALSE)
+    lapply(disp.formulas, function(x) environment(x) <- pf)
+  }else{
+    if(sum(!sapply(disp.formulas, is.null)) != K)
+      stop("Need to supply dispersion formulas for all responses even if not required for all K responses\n",
+           "(You can just fill-in ~1 for those with no wanted dispersion model.)")
+  }
   
   # Necessary parameters & data generation ----
-  if(unif.times){
+  if(regular.times){
     time <- seq(0, fup, length.out = ntms)
     tau <- fup + 0.1
   }else{
@@ -185,25 +267,37 @@ simData <- function(n = 250, ntms = 10, fup = 5,
   cont <- rnorm(n); bin <- rbinom(n, 1, 0.5)  # Continuous and binary covariates.
   
   df <- data.frame(id = rep(1:n, each = ntms),
-                   time = if(unif.times) rep(time, n) else time,
+                   time = if(regular.times) rep(time, n) else time,
                    cont = rep(cont, each = ntms),
                    bin = rep(bin, each = ntms))
   
   # Design matrices, used across ALL responses ----
   X <- model.matrix(~ time + cont + bin, data = df)
+  W <- lapply(disp.formulas, function(x) model.matrix(x, data = df))
   if(!is.null(random.formula)){
     Z <- lapply(random.formula, function(x) model.matrix(x, data = df))
   }else{
     Z <- replicate(K, model.matrix(~ 1 + time, data = df), simplify = F) # assume all intslopes.
   }
+  # Check that the supplied disp.formulas match with elements of sigma
+  if(!all(sapply(1:K, function(k) ncol(W[[k]])==length(sigma[[k]]))))
+    stop("Mis-specified sigma or disp.formulas.\n")
 
+  # Random effects specification
+  if(is.infinite(dof)) 
+    .simRE <- function(n, mu, Sigma) MASS::mvrnorm(n = n, mu = mu, Sigma = Sigma)
+  else
+    .simRE <- function(n, mu, Sigma) mvtnorm::rmvt(n = n, sigma = Sigma, delta = mu, df = dof)
+  
   # Linear predictor & response generation ----
   if(!is.null(random.formula)){
     q <- ncol(do.call(cbind, lapply(Z, head)))
-    b <- MASS::mvrnorm(n, mu = rep(0, q), Sigma = D)
+    # b <- MASS::mvrnorm(n, mu = rep(0, q), Sigma = D)
+    b <- .simRE(n, rep(0, q), D)
     b.inds <- split(1:q, do.call(c, sapply(1:K, function(k) rep(k, ncol(Z[[k]])), simplify = F)))
   }else{
-    b <- MASS::mvrnorm(n, mu = rep(0, K * 2), Sigma = D)
+    # b <- MASS::mvrnorm(n, mu = rep(0, K * 2), Sigma = D)
+    b <- .simRE(n, rep(0, K * 2), D)
     b.inds <- split(1:(2*K), rep(1:K, each = 2)) 
   }
   
@@ -211,22 +305,29 @@ simData <- function(n = 250, ntms = 10, fup = 5,
     f <- family[[k]]; Zk <- Z[[k]]
     betak <- beta[k,,drop=T]
     etak <- X %*% betak + rowSums(Zk * b[df$id, b.inds[[k]]])
+    phik <- W[[k]] %*% sigma[[k]]
     switch(f, 
-           gaussian = Y <- rnorm(n * ntms, mean = etak, sd = sqrt(sigma[k])),
+           gaussian = Y <- rnorm(n * ntms, mean = etak, sd = sqrt(phik[1])),
            binomial = Y <- rbinom(n * ntms, 1, plogis(etak)),
            poisson = Y <- rpois(n * ntms, exp(etak)),
-           Gamma = Y <- rgamma(n * ntms, shape = sigma[k], scale = exp(etak)/sigma[k]),
+           Gamma = Y <- rgamma(n * ntms, shape = exp(phik), scale = exp(etak)/exp(phik)),
+           negbin = Y <- MASS::rnegbin(n * ntms, mu = exp(etak), theta = exp(phik)),
            genpois = {
              maxtry <- 5
              try <- 1
              e <- T
              while(e){ # genpois can be quite sensitive in simulation part; repeat sim up to 5 times.
-               Y <- tryCatch(rgenpois(exp(etak), sigma[k]),
+               Y <- tryCatch(rgenpois(exp(etak), phik),
                              error = function(e) NA)
                e <- any(is.na(Y))
                try <- try + 1
-               if(try > maxtry) stop("Issues creating genpois response with beta values ", c(betak), 
-                                     " and mean(b)", c(round(mean(b[df$id, b.inds[[k]]]), 3)))
+               if(try > maxtry){
+                 stop(sprintf("Issues creating genpois response, this could be due to large values for mu (range: %.2f-%.2f),",
+                              range(exp(etak))[1], range(exp(etak))[2]),
+                      sprintf(" or dispersion (range: %.2f, %.2f); it's recommended that this value is between", 
+                              range(phik)[1], range(phik)[2]),
+                      sprintf(" %.2f and %.2f.\n", max(-1, -max(exp(etak)/4)), 1))
+               }
              }
              Y
            })
@@ -278,7 +379,7 @@ simData <- function(n = 250, ntms = 10, fup = 5,
   
   out.data <- merge(df, surv.data, by = 'id')
   out.data <- out.data[out.data$time < out.data$survtime, ]
-  message(round(100 * sum(surv.data$status)/n), '% failure rate')
+  cat(sprintf("%.2f%% failure rate.\n", 100 * sum(surv.data$status)/n))
   
   out <- list(data =  out.data, 
               surv.data =  out.data[!duplicated(out.data[,'id']), c('id', 'survtime', 'status', 'cont', 'bin')])
